"""
Tarka

:copyright: 2022-2023 Nándor Mátravölgyi
:license: Apache2, see LICENSE for more details.
"""

__copyright__ = "Copyright 2022-2023 Nándor Mátravölgyi"
__credits__ = ["Nándor Mátravölgyi"]
__version__ = "0.15.1"
